//#define VIENNACL_DEBUG_ALL
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cmath>
#include <ctime>
#include <CL/cl.hpp>
#include <petsc.h>
#include <petscviennacl.h>
#include "constants.h"

viennacl::ocl::program program;

extern void InitialCondition(TS ts, Vec prim);
extern PetscErrorCode ComputeResidualViennaCL(TS ts,
                               PetscScalar t,
                               Vec Prim, Vec dPrim_dt,
                               Vec F, void *ptr);
extern PetscErrorCode ComputeResidual(TS ts,
                               PetscScalar t,
                               Vec Prim, Vec dPrim_dt,
                               Vec F, void *ptr);

int main(int argc, char **argv)
{
    TS ts;
    Vec soln;
    DM da, DAs[DOF];
    PetscErrorCode ierr;

    static const char help[] = 
        "Petsc ViennaCL test";
    PetscInitialize(&argc, &argv, PETSC_NULL, help);

//    viennacl::ocl::set_context_platform_index(0, 1);

    ierr = DMCompositeCreate(PETSC_COMM_WORLD, &da); CHKERRQ(ierr);

    for (int var=0; var<DOF; var++) {
        ierr = DMDACreate2d(PETSC_COMM_WORLD, 
                            DMDA_BOUNDARY_GHOSTED, DMDA_BOUNDARY_GHOSTED,
                            DMDA_STENCIL_STAR,
                            N1, N2,
                            PETSC_DECIDE, PETSC_DECIDE,
                            1, NG, PETSC_NULL, PETSC_NULL, &DAs[var]);  CHKERRQ(ierr);

        ierr = DMDASetUniformCoordinates(DAs[var], 
                                  X1_MIN + DX1/2., 
                                  X1_MAX - DX1/2.,
                                  X2_MIN + DX2/2.,
                                  X2_MAX - DX2/2.,
                                  0., 0.); CHKERRQ(ierr);


        ierr = DMCompositeAddDM(da, DAs[var]); CHKERRQ(ierr);
    }

    ierr = DMSetVecType(da, VECVIENNACL); CHKERRQ(ierr);
    ierr = DMSetMatType(da, MATAIJVIENNACL); CHKERRQ(ierr);

    ierr = DMCreateGlobalVector(da, &soln); CHKERRQ(ierr);

    ierr = TSCreate(PETSC_COMM_WORLD, &ts); CHKERRQ(ierr);
    ierr = TSSetDM(ts, da); CHKERRQ(ierr);

//    ierr = TSSetIFunction(ts, PETSC_NULL, ComputeResidual, NULL); CHKERRQ(ierr);
    ierr = TSSetIFunction(ts, PETSC_NULL, ComputeResidualViennaCL, NULL); CHKERRQ(ierr);

    std::ifstream sourceFile("computeresidual.cl");
    std::string sourceCode((std::istreambuf_iterator<char>(sourceFile)),
                            std::istreambuf_iterator<char>());

    PetscScalar start = std::clock();
    program =
        viennacl::ocl::current_context().add_program(sourceCode,
                                                     "computeresidual");
    PetscScalar end = std::clock();
    PetscScalar time = (end - start)/(PetscScalar)CLOCKS_PER_SEC;
    PetscPrintf(PETSC_COMM_WORLD, 
                "Time taken for kernel compilation = %f\n", time);
    
    program.add_kernel("ComputeResidual");

    InitialCondition(ts, soln);

    ierr = TSSetSolution(ts, soln); CHKERRQ(ierr);
    ierr = TSSetFromOptions(ts); CHKERRQ(ierr);

    ierr = TSSolve(ts, soln); CHKERRQ(ierr);

    ierr = DMDestroy(&da); CHKERRQ(ierr);
    for (int var=0; var<DOF; var++) {
        ierr = DMDestroy(&DAs[var]); CHKERRQ(ierr);
    }
    ierr = VecDestroy(&soln); CHKERRQ(ierr);
    ierr = TSDestroy(&ts); CHKERRQ(ierr);


    PetscFinalize();
    return(0);
}

PetscErrorCode ComputeResidualViennaCL(TS ts,
                                       PetscScalar t,
                                       Vec X, Vec dX_dt,
                                       Vec F, void *ptr)
{
    PetscErrorCode ierr;
    const viennacl::vector<PetscScalar> *x, *dx_dt;
    viennacl::vector<PetscScalar> *f;

    ierr = VecViennaCLGetArrayRead(X, &x); CHKERRQ(ierr);
    ierr = VecViennaCLGetArrayRead(dX_dt, &dx_dt); CHKERRQ(ierr);
    ierr = VecViennaCLGetArrayWrite(F, &f); CHKERRQ(ierr);

    viennacl::ocl::kernel kernel = program.get_kernel("ComputeResidual");

    kernel.local_work_size(0, 8);           //FIXED
    kernel.local_work_size(1, 8);           //FIXED
    kernel.global_work_size(0, N1);
    kernel.global_work_size(1, N2);

    viennacl::ocl::enqueue(kernel(*x, *dx_dt, *f));
    viennacl::backend::finish();

    ierr = VecViennaCLRestoreArrayRead(X, &x); CHKERRQ(ierr);
    ierr = VecViennaCLRestoreArrayRead(dX_dt, &dx_dt); CHKERRQ(ierr);
    ierr = VecViennaCLRestoreArrayWrite(F, &f); CHKERRQ(ierr);

    return ierr;           //FIXED
}

PetscErrorCode ComputeResidual(TS ts,
                               PetscScalar t,
                               Vec X, Vec dX_dt,
                               Vec F, void *ptr)
{
    PetscErrorCode ierr;
    PetscScalar *x, *dx_dt, *f;

    ierr = VecGetArray(X, &x); CHKERRQ(ierr);
    ierr = VecGetArray(dX_dt, &dx_dt); CHKERRQ(ierr);
    ierr = VecGetArray(F, &f); CHKERRQ(ierr);
    
    for (int var=0; var<DOF; var++)
        for (int j=0; j<N2-1; j++)
            for (int i=0; i<N1-1; i++) {
                f[INDEX_GLOBAL(i,j,var)] = dx_dt[INDEX_GLOBAL(i,j,var)] -
                                         (x[INDEX_GLOBAL(i+1,j,var)] - 
                                          x[INDEX_GLOBAL(i,j,var)])/DX1 -
                                         (x[INDEX_GLOBAL(i,j+1,var)] - 
                                          x[INDEX_GLOBAL(i,j,var)])/DX2;
            }
    ierr = VecRestoreArray(X, &x); CHKERRQ(ierr);
    ierr = VecRestoreArray(dX_dt, &dx_dt); CHKERRQ(ierr);
    ierr = VecRestoreArray(F, &f); CHKERRQ(ierr);

    return ierr;           //FIXED
}

void InitialCondition(TS ts, Vec X)
{
    PetscScalar *x;

    VecGetArray(X, &x);

    for (PetscInt var=0; var<DOF; var++)
        for (PetscInt j=0; j<N2; j++) {
            for (PetscInt i=0; i<N1; i++) {
            
                PetscScalar X1Coord = X1_MIN + DX1/2. + i*DX1;
                PetscScalar X2Coord = X2_MIN + DX2/2. + j*DX2;

                PetscScalar X1Center = (X1_MIN + X1_MAX)/2.;
                PetscScalar X2Center = (X2_MIN + X2_MAX)/2.;

                PetscScalar r = PetscSqrtScalar(
                                    PetscPowScalar(X1Coord-X1Center, 2.0) + 
                                    PetscPowScalar(X2Coord-X2Center, 2.0));

                x[i + N1*(j + N2*var)] = exp(-r*r/.01);

            }
        }

    VecRestoreArray(X, &x);
}

